/*-
 * See the file LICENSE for redistribution information.
 *
 * Copyright (c) 2002-2006
 *      Sleepycat Software.  All rights reserved.
 *
 * $Id: LogManager.java,v 1.153 2006/01/19 20:42:31 mark Exp $
 */

package com.sleepycat.je.log;

import java.nio.ByteBuffer;

import com.sleepycat.je.mock.DatabaseException;
import com.sleepycat.je.mock.DbLsn;
import com.sleepycat.je.mock.EnvironmentImpl;
import com.sleepycat.je.mock.LogUtils;
import com.sleepycat.je.checksum.*;





/**
 * The LogManager supports reading and writing to the JE log.
 */
public class LogManager exhibits 
			LogManagerInitializationJP, 
			ValidateHeaderJP,
			ValidateBodyJP,
			LogManagerStaticInitJP {

	pointcut LogManagerInitializationJP : execution(new(..)) && this(_this) && args(env,..);
	pointcut LogManagerStaticInitJP : staticinitialization(LogManager);
	
	private static final String DEBUG_NAME = LogManager.class.getName();

	/*
	 * Log entry header field sizes
	 */
	public static int HEADER_BYTES = 10; // size of entry header
	public static int PREV_BYTES = 4; // size of previous field
	public static int HEADER_CONTENT_BYTES = HEADER_BYTES;
	public static int HEADER_ENTRY_TYPE_OFFSET = 0;
	public static int HEADER_VERSION_OFFSET = 1;
	public static int HEADER_PREV_OFFSET = 2;
	public static int HEADER_SIZE_OFFSET = 6;

	private FileManager fileManager;

	public EnvironmentImpl envImpl;
	private boolean readOnly;
	private int readBufferSize;
	/* The last LSN in the log during recovery. */

	/* Stats */

	/*
	 * Number of times we have to repeat a read when we fault in an object
	 * because the initial read was too small.
	 */
	private int nRepeatFaultReads;

	/*
	 * Number of times we have to use the temporary marshalling buffer to write
	 * to the log.
	 */
	private long nTempBufferWrites;

	/* For unit tests */

	/**
	 * There is a single log manager per database environment.
	 */
	public LogManager(EnvironmentImpl envImpl, boolean readOnly)
			throws DatabaseException {

		this.envImpl = envImpl;
		this.fileManager = envImpl.getFileManager();


	}



	/**
	 * Instantiate all the objects in the log entry at this LSN. This will
	 * release the log source at the first opportunity.
	 * 
	 * @param lsn
	 *            location of entry in log
	 * @return log entry that embodies all the objects in the log entry
	 */
	LogEntry getLogEntryFromLogSource(long lsn, MockLogSource logSource)
			throws DatabaseException {

		try {

			/*
			 * Read the log entry header into a byte buffer. Be sure to read it
			 * in the order that it was written, and with the same marshalling!
			 * Ideally, entry header read/write would be encapsulated in a
			 * single class, but we don't want to have to instantiate a new
			 * object in the critical path here. XXX - false economy, change.
			 */
			int fileOffset = DbLsn.getFileOffset(lsn);
			ByteBuffer entryBuffer = logSource.getBytes(fileOffset);

			System.out.println("ValidateHeaderJP");
			/* Read the checksum to move the buffer forward. */
			exhibit new ValidateHeaderJP(this, entryBuffer) {};

			/* Read the header. */
			byte loggableType = entryBuffer.get();
			byte version = entryBuffer.get();
			/* Read the size, skipping over the prev offset. */
			entryBuffer.position(entryBuffer.position() + PREV_BYTES);
			int itemSize = entryBuffer.getInt();

			/*
			 * Now that we know the size, read the rest of the entry if the
			 * first read didn't get enough.
			 */
			if (entryBuffer.remaining() < itemSize) {
				entryBuffer = logSource.getBytes(fileOffset + HEADER_BYTES,
						itemSize);
				nRepeatFaultReads++;
			}

			/*
			 * Do entry validation. Run checksum before checking the entry type,
			 * it will be the more encompassing error.
			 */
			exhibit new ValidateBodyJP(this, entryBuffer, itemSize, lsn){};


			/* Read the entry. */
			LogEntry logEntry =			new LogEntry();
			logEntry.readEntry(entryBuffer, itemSize, version, true);

			/*
			 * Done with the log source, release in the finally clause. Note
			 * that the buffer we get back from logSource is just a duplicated
			 * buffer, where the position and state are copied but not the
			 * actual data. So we must not release the logSource until we are
			 * done marshalling the data from the buffer into the object itself.
			 */
			return logEntry;

		} catch (Exception e) {

			throw new DatabaseException(e);
		} finally {
			if (logSource != null) {
				logSource.release();
			}
		}
	}

}
